from returns.result import Result, Success, Failure
from returns.pipeline import is_successful

from .experiment import Experiment
from .pipeline import Pipeline
from .task import Task, TaskDispatcher



from __future__ import annotations

from copy import deepcopy

from itertools import cycle, chain
from uuid import uuid4
from typing import Dict, Set, Union, Iterator, Callable, Sequence, Any, List

import netunicorn
from .architecture import Architecture
from .types import NodeProperty, NodeRepresentation, NodesRepresentation

from abc import ABC, abstractmethod


class Node:
    def __init__(
        self,
        name: str,
        properties: Dict[str, NodeProperty],
        architecture: Architecture = Architecture.UNKNOWN,
    ):
        self.name = name
        self.properties = properties
        self.additional_properties: Dict[str, NodeProperty] = {}
        self.architecture = architecture

    def __getitem__(self, item: str) -> NodeProperty:
        return self.properties.get(item, None)

    def __setitem__(self, key: str, value: NodeProperty) -> None:
        self.properties[key] = value

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Node)
            and self.name == other.name
            and self.properties == other.properties
            and self.additional_properties == other.additional_properties
            and self.architecture == other.architecture
        )

    def __json__(self) -> NodeRepresentation:
        return {
            "name": self.name,
            "properties": self.properties,
            "additional_properties": self.additional_properties,
            "architecture": self.architecture.value,
        }

    @classmethod
    def from_json(cls, data: NodeRepresentation) -> Node:
        instance = cls(
            data["name"], data["properties"], Architecture(data["architecture"])
        )
        instance.additional_properties = data["additional_properties"]
        return instance


class Nodes(ABC):
    """
    Represents a pool of nodes.
    """

    @abstractmethod
    def __json__(self) -> NodesRepresentation:
        """
        Returns a JSON representation of the object.
        """
        pass

    @staticmethod
    def dispatch_and_deserialize(data: NodesRepresentation) -> Nodes:
        cls: Nodes = getattr(netunicorn.base.nodes, data["node_pool_type"])
        return cls.from_json(data["node_pool_data"])

    @classmethod
    @abstractmethod
    def from_json(
        cls, data: List[Union[NodeRepresentation, NodesRepresentation]]
    ) -> Nodes:
        """
        Accepts a JSON representation of the object (from "node_pool_data") and returns an instance of the object.
        """
        pass

    @abstractmethod
    def __str__(self) -> str:
        """
        Returns a string representation of the object.
        """
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, item: int) -> Union[Node, Nodes]:
        pass

    @abstractmethod
    def filter(self, function: Callable[[Node], bool]) -> Nodes:
        """
        Returns a pool of nodes that match the given filter.
        """
        pass

    @abstractmethod
    def take(self, count: int) -> Sequence[Node]:
        """
        Returns a sequence of nodes consisting of the first n nodes.
        """
        pass

    @abstractmethod
    def skip(self, count: int) -> Nodes:
        """
        Returns a pool of nodes consisting of the nodes after the first n nodes.
        """
        pass

    @abstractmethod
    def set_property(self, name: str, value: NodeProperty) -> Nodes:
        pass


class CountableNodePool(Nodes):
    """
    Represents a typical pool of nodes.
    """

    def __init__(self, nodes: List[Union[Node, Nodes]]):
        self.nodes = nodes

    def __json__(self) -> NodesRepresentation:
        return {
            "node_pool_type": self.__class__.__name__,
            "node_pool_data": [x.__json__() for x in self.nodes],
        }

    @classmethod
    def from_json(
        cls, data: List[Union[NodeRepresentation, NodesRepresentation]]
    ) -> CountableNodePool:
        nodes: List[Union[Node, Nodes]] = []
        for element in data:
            if "node_pool_type" in element:
                # we know this is a NodesRepresentation
                nodes.append(Nodes.dispatch_and_deserialize(element))  # type: ignore
            else:
                nodes.append(Node.from_json(element))
        return cls(nodes)

    def __str__(self) -> str:
        return str(self.nodes)

    def __len__(self) -> int:
        return len(self.nodes)

    def __getitem__(self, key: int) -> Union[Node, Nodes]:
        return self.nodes[key]

    def __setitem__(
        self, key: int, value: Union[Node, CountableNodePool, UncountableNodePool]
    ) -> None:
        self.nodes[key] = value

    def pop(self, index: int) -> Union[Node, Nodes]:
        return self.nodes.pop(index)

    def __repr__(self) -> str:
        return str(self.nodes)

    def filter(self, function: Callable[[Node], bool]) -> CountableNodePool:
        nodes: List[Union[Node, Nodes]] = []
        for node in self.nodes:
            if isinstance(node, Node):
                if function(node):
                    nodes.append(node)
            else:
                filtered_nodes = node.filter(function)
                if len(filtered_nodes) > 0:
                    nodes.append(filtered_nodes)
        return CountableNodePool(nodes)

    def take(self, count: int) -> Sequence[Node]:
        iterator: Iterator[Node] = chain.from_iterable(
            [x] if isinstance(x, Node) else x for x in self.nodes  # type: ignore
        )
        nodes = []
        for _ in range(count):
            try:
                nodes.append(next(iterator))
            except StopIteration:
                break
        return nodes

    def skip(self, count: int) -> CountableNodePool:
        if count > len(self.nodes):
            print(
                f"Warning: asked to skip {count} nodes, but pool length is {len(self.nodes)}. "
                f"Returning empty pool."
            )
            return CountableNodePool([])
        return CountableNodePool(self.nodes[count:])

    def set_property(self, name: str, value: NodeProperty) -> CountableNodePool:
        for node in self.nodes:
            if isinstance(node, Node):
                node.properties[name] = value
            else:
                node.set_property(name, value)
        return self


class UncountableNodePool(Nodes):
    """
    Represents a pool of nodes that is not countable (e.g., dynamic pools of cloud providers).
    In the current implementation cannot have Nodes as elements.
    """

    def __init__(self, node_template: List[Node]):
        self._node_template = node_template
        self._nodes = cycle(node_template)

    def __str__(self) -> str:
        return str(
            f"<Uncountable node pool with next node template: {self._node_template}>"
        )

    def __repr__(self) -> str:
        return str(self)

    def __next__(self) -> Node:
        node = deepcopy(next(self._nodes))
        node.name += str(uuid4())
        return node

    def __getitem__(self, key: int) -> Node:
        return self._node_template[key]

    def __setitem__(self, key: int, value: Node) -> None:
        self._node_template[key] = value

    def __len__(self) -> int:
        return len(self._node_template)

    def filter(self, function: Callable[[Node], bool]) -> UncountableNodePool:
        return UncountableNodePool([x for x in self._node_template if function(x)])

    def take(self, count: int) -> Sequence[Node]:
        nodes = []
        for _ in range(count):
            nodes.append(next(self))
        return nodes

    def skip(self, count: int) -> UncountableNodePool:
        count = count % len(self._node_template)
        for _ in range(count):
            next(self._nodes)
        return self

    def __json__(self) -> NodesRepresentation:
        return {
            "node_pool_type": self.__class__.__name__,
            "node_pool_data": [x.__json__() for x in self._node_template],
        }
        pass

    @classmethod
    def from_json(
        cls, data: List[Union[NodeRepresentation, NodesRepresentation]]
    ) -> UncountableNodePool:
        for x in data:
            if "node_pool_type" in x:
                raise ValueError("UncountableNodePool cannot have Nodes as elements.")
        # now we have only NodeRepresentation in the list
        return cls([Node.from_json(x) for x in data])  # type: ignore

    def set_property(self, name: str, value: NodeProperty) -> UncountableNodePool:
        for node in self._node_template:
            node.properties[name] = value
        return self



from __future__ import annotations

import platform
from dataclasses import dataclass, field
from typing import Optional, Dict, List

from .types import (
    RuntimeContextRepresentation,
    EnvironmentDefinitionRepresentation,
    ShellExecutionRepresentation,
    BuildContextRepresentation,
    DockerImageRepresentation,
)


# TODO: make classes frozen to make hash stable


@dataclass
class RuntimeContext:
    ports_mapping: Dict[int, int] = field(default_factory=dict)
    environment_variables: Dict[str, str] = field(default_factory=dict)
    additional_arguments: List[str] = field(default_factory=list)

    def __json__(self) -> RuntimeContextRepresentation:
        return {
            "ports_mapping": self.ports_mapping,
            "environment_variables": self.environment_variables,
            "additional_arguments": self.additional_arguments,
        }

    @classmethod
    def from_json(cls, data: RuntimeContextRepresentation) -> RuntimeContext:
        ports_mapping = data["ports_mapping"]
        environment_variables = data["environment_variables"]
        additional_arguments = data["additional_arguments"]
        return cls(
            ports_mapping=ports_mapping,
            environment_variables=environment_variables,
            additional_arguments=additional_arguments,
        )


@dataclass
class EnvironmentDefinition:
    commands: List[str] = field(default_factory=list)

    def __hash__(self) -> int:
        return hash(tuple(self.commands))

    @classmethod
    def from_json(
        cls, data: EnvironmentDefinitionRepresentation
    ) -> EnvironmentDefinition:
        instance = cls.__new__(cls)
        instance.commands = data["commands"]
        return instance

    def __json__(self) -> EnvironmentDefinitionRepresentation:
        return {
            "commands": self.commands,
        }


@dataclass
class ShellExecution(EnvironmentDefinition):
    """
    This class represents Environment Definition that's created by executing shell commands.
    """

    runtime_context: RuntimeContext = field(default_factory=RuntimeContext)

    def __json__(self) -> ShellExecutionRepresentation:
        return {
            "commands": self.commands,
            "runtime_context": self.runtime_context.__json__(),
        }

    @classmethod
    def from_json(cls, _data: EnvironmentDefinitionRepresentation) -> ShellExecution:
        if "runtime_context" not in _data:
            raise ValueError("runtime_context is missing")
        data: ShellExecutionRepresentation = _data  # type: ignore
        instance = cls.__new__(cls)
        instance.commands = data["commands"]
        instance.runtime_context = RuntimeContext.from_json(data["runtime_context"])
        return instance


@dataclass(frozen=True)
class BuildContext:
    """ """

    python_version: str = field(default_factory=lambda: platform.python_version())
    cloudpickle_version: Optional[str] = field(
        default_factory=lambda: BuildContext._get_cloudpickle_version()
    )

    @staticmethod
    def _get_cloudpickle_version() -> Optional[str]:
        try:
            import cloudpickle

            return cloudpickle.__version__  # type: ignore
        except ImportError:
            return None

    def __json__(self) -> BuildContextRepresentation:
        return {
            "python_version": self.python_version,
            "cloudpickle_version": self.cloudpickle_version,
        }

    @classmethod
    def from_json(cls, data: BuildContextRepresentation) -> BuildContext:
        python_version = data["python_version"]
        cloudpickle_version = data["cloudpickle_version"]
        return cls(
            python_version=python_version, cloudpickle_version=cloudpickle_version
        )


@dataclass
class DockerImage(EnvironmentDefinition):
    """
    This class represents Environment Definition that is created by using a Docker image.
    If image name is not provided, then it would be created automatically
    """

    image: Optional[str] = None
    build_context: BuildContext = field(default_factory=BuildContext)
    runtime_context: RuntimeContext = field(default_factory=RuntimeContext)

    def __hash__(self) -> int:
        if self.image:
            return hash(self.image)

        return hash(
            (
                self.build_context.python_version,
                self.build_context.cloudpickle_version,
                tuple(self.commands),
            )
        )

    def __json__(self) -> DockerImageRepresentation:
        return {
            "image": self.image,
            "build_context": self.build_context.__json__(),
            "runtime_context": self.runtime_context.__json__(),
            "commands": self.commands,
        }

    @classmethod
    def from_json(cls, _data: EnvironmentDefinitionRepresentation) -> DockerImage:
        if "image" not in _data:
            raise ValueError("image is missing")
        data: DockerImageRepresentation = _data  # type: ignore
        instance = cls.__new__(cls)
        instance.commands = data["commands"]
        instance.image = data["image"]
        instance.build_context = BuildContext.from_json(data["build_context"])
        instance.runtime_context = RuntimeContext.from_json(data["runtime_context"])
        return instance



from __future__ import annotations
import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from typing import Dict, Set, Union, List, Optional

NodeProperty = Union[str, float, int, Set[str], None]


class DeploymentRepresentation(TypedDict):
    node: NodeRepresentation
    prepared: bool
    executor_id: str
    error: Optional[str]
    pipeline: bytes
    keep_alive_timeout_minutes: int
    environment_definition: EnvironmentDefinitionRepresentation
    environment_definition_type: str


class NodeRepresentation(TypedDict):
    name: str
    properties: Dict[str, NodeProperty]
    additional_properties: Dict[str, NodeProperty]
    architecture: str


class NodesRepresentation(TypedDict):
    node_pool_type: str
    node_pool_data: List[Union[NodeRepresentation, NodesRepresentation]]


class RuntimeContextRepresentation(TypedDict):
    ports_mapping: Dict[int, int]
    environment_variables: Dict[str, str]
    additional_arguments: List[str]


class EnvironmentDefinitionRepresentation(TypedDict):
    commands: List[str]


class ShellExecutionRepresentation(EnvironmentDefinitionRepresentation):
    runtime_context: RuntimeContextRepresentation


class DockerImageRepresentation(EnvironmentDefinitionRepresentation):
    image: Optional[str]
    build_context: BuildContextRepresentation
    runtime_context: RuntimeContextRepresentation


class BuildContextRepresentation(TypedDict):
    python_version: str
    cloudpickle_version: Optional[str]


class ExperimentRepresentation(TypedDict):
    deployment_map: List[DeploymentRepresentation]


class DeploymentExecutionResultRepresentation(TypedDict):
    node: NodeRepresentation
    pipeline: str
    result: Optional[str]
    error: Optional[str]


class ExperimentExecutionInformationRepresentation(TypedDict):
    status: int
    experiment: Optional[ExperimentRepresentation]
    execution_result: Union[None, str, List[DeploymentExecutionResultRepresentation]]



import dataclasses
import multiprocessing.pool
from base64 import b64encode
from functools import wraps
from json import JSONEncoder
from typing import Callable, List, TypeVar, Union, Any, Dict

from netunicorn.base.environment_definitions import EnvironmentDefinition
from returns.result import Failure, Result, Success

_ValueType = TypeVar("_ValueType", covariant=True)
_FailureValueType = TypeVar("_FailureValueType", covariant=True)

_FunctionType = Union[
    Callable[..., _ValueType], Callable[..., Result[_ValueType, _FailureValueType]]
]

SerializedPipelineType = bytes
LogType = List[str]


def safe(
    function: _FunctionType,  # type: ignore
) -> Union[
    Callable[..., Result[_ValueType, Exception]],
    Callable[..., Result[_ValueType, _FailureValueType]],
]:
    """
    Decorator that wraps function in try/except block.
    :param function: function to wrap
    :return: wrapped function
    """

    @wraps(function)
    def decorator(*args: Any, **kwargs: Any) -> Result[Any, Any]:
        try:
            result = function(*args, **kwargs)
            if result is None:
                """
                Well, ok. The problem is that if we create Success(None) object,
                cloudpickle serialization somewhere breaks and we receive this error:
                AttributeError: 'Success' object has no attribute '_inner_value'
                So, for now - we return 0 instead of None until we find and fix/report the bug.
                """
                result = 0
            if isinstance(result, Result):
                return result
            return Success(result)
        except Exception as exc:
            return Failure(exc.__reduce__())

    return decorator


# If you try to run multiprocessing.Pool inside a multiprocessing.Pool, you'll receive this error:
# AssertionError: daemonic processes are not allowed to have children
# Below is a solution to this problem.
class NoDaemonProcess(multiprocessing.Process):
    @property
    def daemon(self) -> bool:
        return False

    @daemon.setter
    def daemon(self, value: Any) -> None:
        pass


class NoDaemonContext(type(multiprocessing.get_context())):  # type: ignore
    Process = NoDaemonProcess


# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.
class NonStablePool(multiprocessing.pool.Pool):
    # noinspection PyArgumentList
    def __init__(self, *args: Any, **kwargs: Any):
        kwargs["context"] = NoDaemonContext()
        super().__init__(*args, **kwargs)


class UnicornEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, Exception):
            return str(obj.__reduce__())
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        if hasattr(obj, "__json__"):
            return obj.__json__()
        if isinstance(obj, bytes):
            return b64encode(obj).decode("utf-8")
        if isinstance(obj, Result):
            # noinspection PyProtectedMember
            return {"result_type": obj.__class__.__name__, "result": obj._inner_value}
        if isinstance(obj, EnvironmentDefinition):
            return {
                "environment_definition_type": obj.__class__.__name__,
                "environment_definition": obj,
            }
        return JSONEncoder.default(self, obj)



from __future__ import annotations

from base64 import b64decode
from typing import Optional
from copy import deepcopy

import netunicorn.base.environment_definitions

from .nodes import Node
from .pipeline import Pipeline
from .task import TaskDispatcher, Task
from .utils import SerializedPipelineType
from .types import DeploymentRepresentation

try:
    import cloudpickle  # it's needed only on client side, but this module is also imported on engine side
    import netunicorn.library

    cloudpickle.register_pickle_by_value(netunicorn.library)
except ImportError:
    pass


class Deployment:
    def __init__(
        self, node: Node, pipeline: Pipeline, keep_alive_timeout_minutes: int = 10
    ):
        self.node = node
        self.prepared = False
        self.executor_id = ""
        self.error: Optional[Exception] = None
        self.pipeline: SerializedPipelineType = b""
        self.environment_definition = deepcopy(pipeline.environment_definition)
        self.keep_alive_timeout_minutes = keep_alive_timeout_minutes

        pipeline = deepcopy(pipeline)

        for i, element in enumerate(pipeline.tasks):
            pipeline.tasks[i] = [
                x.dispatch(node) if isinstance(x, TaskDispatcher) else x
                for x in element
            ]
            for x in pipeline.tasks[i]:
                # now it's only Tasks
                self.environment_definition.commands.extend(x.requirements)  # type: ignore

        self.pipeline = cloudpickle.dumps(pipeline)

    def __str__(self) -> str:
        return f"Deployment: Node={self.node.name}, executor_id={self.executor_id}, prepared={self.prepared}"

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> DeploymentRepresentation:
        return {
            "node": self.node.__json__(),
            "prepared": self.prepared,
            "executor_id": self.executor_id,
            "error": str(self.error) if self.error else None,
            "pipeline": self.pipeline,
            "keep_alive_timeout_minutes": self.keep_alive_timeout_minutes,
            "environment_definition": self.environment_definition.__json__(),
            "environment_definition_type": self.environment_definition.__class__.__name__,
        }

    @classmethod
    def from_json(cls, data: DeploymentRepresentation) -> Deployment:
        instance = cls.__new__(cls)

        instance.node = Node.from_json(data["node"])
        instance.prepared = data["prepared"]
        instance.executor_id = data["executor_id"]
        instance.error = Exception(data["error"]) if data["error"] else None
        instance.pipeline = b64decode(data["pipeline"])
        instance.keep_alive_timeout_minutes = data["keep_alive_timeout_minutes"]
        instance.environment_definition = getattr(
            netunicorn.base.environment_definitions, data["environment_definition_type"]
        ).from_json(data["environment_definition"])
        return instance



import json

import asyncpg

from netunicorn.base.utils import UnicornEncoder


async def __init_connection(conn: asyncpg.Connection) -> None:
    await conn.set_type_codec(
        "jsonb",
        encoder=lambda x: json.dumps(x, cls=UnicornEncoder),
        decoder=json.loads,
        schema="pg_catalog",
    )



from __future__ import annotations

import base64
import copy
from dataclasses import dataclass
from enum import Enum
from typing import Iterator, List, Optional, Tuple, Union, Sequence

from returns.result import Result

from .deployment import Deployment
from .nodes import Node, Nodes
from .pipeline import Pipeline, PipelineResult
from .types import (
    ExperimentRepresentation,
    DeploymentExecutionResultRepresentation,
    ExperimentExecutionInformationRepresentation,
)
from .utils import LogType


class ExperimentStatus(Enum):
    UNKNOWN = 0
    PREPARING = 1
    READY = 2
    RUNNING = 3
    FINISHED = 4

    def __json__(self) -> int:
        return self.value

    @classmethod
    def from_json(cls, value: int) -> ExperimentStatus:
        return cls(value)


class Experiment:
    def __init__(self) -> None:
        self.deployment_map: List[Deployment] = []

    def append(self, node: Node, pipeline: Pipeline) -> Experiment:
        self.deployment_map.append(Deployment(node, pipeline))
        return self

    def map(self, nodes: Sequence[Node], pipeline: Pipeline) -> Experiment:
        if isinstance(nodes, Nodes):
            raise TypeError("Expected sequence of nodes, got Nodes instead")

        for node in nodes:
            self.append(node, pipeline)
        return self

    def __json__(self) -> ExperimentRepresentation:
        return {
            "deployment_map": [x.__json__() for x in self.deployment_map],
        }

    @classmethod
    def from_json(cls, data: ExperimentRepresentation) -> Experiment:
        instance = cls.__new__(cls)
        instance.deployment_map = [
            Deployment.from_json(x) for x in data["deployment_map"]
        ]
        return instance

    def __getitem__(self, item: int) -> Deployment:
        return self.deployment_map[item]

    def __iter__(self) -> Iterator[Deployment]:
        return iter(self.deployment_map)

    def __len__(self) -> int:
        return len(self.deployment_map)

    def __str__(self) -> str:
        return "; ".join([f"<{x}>" for x in self.deployment_map])

    def __repr__(self) -> str:
        return str(self)

    def __add__(self, other: Experiment) -> Experiment:
        new_map = copy.deepcopy(self)
        new_map.deployment_map.extend(other.deployment_map)
        return new_map


class DeploymentExecutionResult:
    def __init__(
        self,
        node: Node,
        serialized_pipeline: bytes,
        result: Optional[bytes],
        error: Optional[str] = None,
    ):
        self.node = node
        self._pipeline = serialized_pipeline
        self._result = result
        self.error = error

    @property
    def pipeline(self) -> Pipeline:
        import cloudpickle

        return cloudpickle.loads(self._pipeline)  # type: ignore

    @property
    def result(
        self,
    ) -> Optional[Tuple[Result[PipelineResult, PipelineResult], LogType]]:
        import cloudpickle

        return cloudpickle.loads(self._result) if self._result else None

    def __str__(self) -> str:
        return f"DeploymentExecutionResult(node={self.node}, result={self.result}, error={self.error})"

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> DeploymentExecutionResultRepresentation:
        return {
            "node": self.node.__json__(),
            "pipeline": base64.b64encode(self._pipeline).decode("utf-8"),
            "result": base64.b64encode(self._result).decode("utf-8")
            if self._result
            else None,
            "error": self.error,
        }

    @classmethod
    def from_json(
        cls, data: DeploymentExecutionResultRepresentation
    ) -> DeploymentExecutionResult:
        return cls(
            Node.from_json(data["node"]),
            base64.b64decode(data["pipeline"]),
            base64.b64decode(data["result"]) if data["result"] else None,
            data["error"],
        )


@dataclass(frozen=True)
class ExperimentExecutionInformation:
    status: ExperimentStatus
    experiment: Optional[Experiment]
    execution_result: Union[None, Exception, List[DeploymentExecutionResult]]

    def __json__(self) -> ExperimentExecutionInformationRepresentation:
        execution_result: Union[
            None, str, List[DeploymentExecutionResultRepresentation]
        ] = None
        if isinstance(self.execution_result, list):
            execution_result = [x.__json__() for x in self.execution_result]
        elif isinstance(self.execution_result, Exception):
            execution_result = str(self.execution_result.__reduce__())
        return {
            "status": self.status.__json__(),
            "experiment": self.experiment.__json__() if self.experiment else None,
            "execution_result": execution_result,
        }

    @classmethod
    def from_json(
        cls, data: ExperimentExecutionInformationRepresentation
    ) -> ExperimentExecutionInformation:
        status = ExperimentStatus.from_json(data["status"])
        experiment = (
            Experiment.from_json(data["experiment"]) if data["experiment"] else None
        )
        execution_result_data = data["execution_result"]
        execution_result: Union[None, Exception, List[DeploymentExecutionResult]] = None
        if execution_result_data:
            if isinstance(execution_result_data, list):
                execution_result = [
                    DeploymentExecutionResult.from_json(x)
                    for x in execution_result_data
                ]
            else:
                execution_result = Exception(execution_result_data)
        return cls(status, experiment, execution_result)



from __future__ import annotations

import uuid
from copy import deepcopy
from typing import Collection, List, Union, Any, Optional

from returns.result import Result

from .environment_definitions import DockerImage, EnvironmentDefinition
from .task import Task, TaskDispatcher

TaskElement = Union[Task, TaskDispatcher]
PipelineElement = Union[TaskElement, Collection[TaskElement]]
PipelineElementResult = Union[Result[Any, Any], Collection[Result[Any, Any]]]
PipelineResult = Collection[PipelineElementResult]


class Pipeline:
    """
    Pipeline is a class that takes a tuple of Tasks and executes them in order.
    Each element in the tuple should be either a Task or a tuple of Tasks.
    Pipeline will execute elements in order and return the combined result of all tasks.
    If element is a tuple of tasks, these tasks would be executed in parallel.

    The result of pipeline execution would be one of the following:
    - Success: if all tasks succeed, then Success is returned
    - Failure: if any task fails, then Failure is returned

    Returning object will always contain a Result object for each task executed.
    If early_stopping is set to True, then any task after first failed wouldn't be executed.
    """

    def __init__(
        self,
        tasks: Collection[PipelineElement] = (),
        early_stopping: bool = True,
        report_results: bool = True,
        environment_definition: Optional[EnvironmentDefinition] = None,
    ):
        """
        Initialize Pipeline with a tuple of Tasks or TaskDispatchers and early_stopping flag.
        :param tasks: a tuple of tasks to be executed
        :param early_stopping: whether to stop executing tasks after first failure
        :param report_results: whether executor should connect director services to report pipeline results after exec
        """
        self.name = str(uuid.uuid4())
        self.early_stopping = early_stopping
        self.tasks: List[List[TaskElement]] = []
        self.report_results = report_results
        self.environment_definition = environment_definition or DockerImage()
        for element in tasks:
            self.then(element)

    @staticmethod
    def __element_to_stage(element: PipelineElement) -> List[TaskElement]:
        if not isinstance(element, Collection):
            element = [element]
        if not isinstance(element, list):
            element = list(element)
        return element

    def then(self, element: PipelineElement) -> Pipeline:
        """
        Add a task or list of tasks to the end of the pipeline.
        :param element: a task or tuple of tasks to be added
        :return: self
        """
        element = self.__element_to_stage(element)
        self.tasks.append(element)
        return self

    def copy(self) -> Pipeline:
        """
        Return a copy of the pipeline.
        :return: a copy of the pipeline
        """
        return Pipeline(deepcopy(self.tasks), self.early_stopping)

    def __str__(self) -> str:
        return f"Pipeline({self.name}): {self.tasks}"

    def __repr__(self) -> str:
        return str(self)
